<?php

// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
/**
 * LICENSE:
 * 
 * Paloose is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program.  If not, see <http://www.gnu.org/licenses/> 
 *
 * Concepts of the software are Copyright 1999-2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 ( the "License" );
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 *   
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @package paloose
 * @subpackage matching
 * @author Hugh Field-Richards <hsfr@hsfr.org.uk>
 * @version See {@link Paloose.php}
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @copyright 2006 - 2011 Hugh Field-Richards, Apache Software Foundation (see below)
 *
 * I freely admit that I have plagiarised most of the following from the Apache
 * Cocoon WildcardHelper package on which it is based.
 *
 */

//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
/**
 */

require_once( PALOOSE_LIB_DIRECTORY . "/exception-handling/MatcherException.php" );

// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* -->
//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
/**
 * The <i>RegexpURIMatcher</i> supports regular expression matching.
 *
 * For example take the wildcard matcher
 *
 * <pre>   &lt;map:match pattern="* /*.html">
 *       ...
 *   &lt;/map:match></pre>
 *
 * The wildcards in the match pattern are assigned to internal pattern variables "{1}", "{2}"
 * etc, in a similar fashion to Perl regular expressions. So in the above case if the
 * pattern to be matched is "documentation/index.html" then there would be a match.
 * The internal variables would be set {1} = "documentation" and {2} = "index". These
 * variables are used within the various pipeline components to resolve string expressions.
 * For example
 * 
 * <pre>   &lt;map:match pattern="* /*.html">
 *       &lt;map:generate src="context://{1}/{2}.xml"/>
 *       ...
 *    &lt;/map:match></pre>
 *
 * In this case the generator would fetch the file documentation/index.xml from the current
 * application context (see pseudo protocols below). Note that the constructions "**" and "*"
 * are different although identical to their Cocoon counterparts.
 * 
 * <ul>
 *   <li>"*"  - Match any character except the path separator.</li>
 *   <li>"**"  - Match any character including the path separator.</li>
 * </ul>
 *
 * The values for the pattern variables are taken from the current matcher. If you want to
 * access a higher matcher (wherever it is) then you would have to use the string {../1},
 * which means access the pattern variable {1} in the previous matcher. These can be extended
 * as necessary, for example {../../1} to access a matcher two levels above the current one.
 *
 * @package paloose
 * @subpackage matching
 */

class WildcardURIMatcher {

   /** The int representing '*' in the pattern int [ ]. */
   const MATCH_FILE = -1;
   /** The int representing '**' in the pattern int [ ]. */
   const MATCH_PATH = -2;
   /** The int representing begin in the pattern int [ ]. */
   const MATCH_BEGIN = -4;
   /** The int representing end in pattern int [ ]. */
   const MATCH_THEEND = -5;
   /** The int value that terminates the pattern int [ ]. */
   const MATCH_END = -3;
    
   private $gHashMap;
    
   /** Logger instance for this class */   
   private $gLogger;

   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Construct a new instance of <i>WildcardURIMatcher</i>.
    */

   public function __construct()
   {
        $this->gLogger = Logger::getLogger( __CLASS__ );
        $this->gHashMap = array();
   }

   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * This function translates a String into an int array
    * converting the special '*' and '\' characters.
    * <br>
    * Here is how the conversion algorithm works:
    * <ul>
    *   <li>The '*' character is converted to MATCH_FILE, meaning that zero
    *        or more characters ( excluding the path separator '/' ) are to
    *        be matched.</li>
    *   <li>The '**' sequence is converted to MATCH_PATH, meaning that zero
    *       or more characters ( including the path separator '/' ) are to
    *        be matched.</li>
    *   <li>The '\' character is used as an escape sequence ( '\*' is
    *       translated in '*', not in MATCH_FILE ). If an exact '\' character
    *       is to be matched the source string must contain a '\\'.
    *       sequence.</li>
    * </ul>
    * When more than two '*' characters, not separated by another character,
    * are found their value is considered as '**' ( MATCH_PATH ).
    * <br>
    * The array is always terminated by a special value ( MATCH_END ).
    * <br>
    * All MATCH* values are less than zero, while normal characters are equal
    * or greater.
    *
    * @param string $inData The string to translate.
    * @retval array The encoded string as an int array, terminated by the MATCH_END
    *         value ( don't consider the array length ).
    */

   private function compilePattern( $inData )
   {
   	$expr = "";
      unset( $expr );
      $buff = $this->toCharArray( $inData );
      
      //Prepare variables for the translation loop
      $y = 0;
      $slash = false;
      
      //Must start from beginning
      $expr[ $y++ ] = self::MATCH_BEGIN;
      
      if ( !empty( $buff ) ) {
         if ( $buff[ 0 ]=='\\' ) {
             $slash = true;
         } else if ( $buff[ 0 ] == '*' ) {
             $expr[ $y++ ] = self::MATCH_FILE;
         }  else {
             $expr[ $y++ ] = $buff[ 0 ];
         }
      
         //Main translation loop
         for ( $x = 1; $x < count( $buff ); $x++ ) {
             // If the previous char was '\' simply copy this char.
             if ( $slash ) {
                 $expr[ $y++ ] = $buff[ $x ];
                 $slash = false;
             // If the previous char was not '\' we have to do a bunch of checks
             } else {
                 // If this char is '\' declare that and continue
                 if ( $buff[ $x ] == '\\' ) {
                     $slash = true;
                 // If this char is '*' check the previous one
                 } else if ( $buff[ $x ] == '*' ) {
                     // If the previous character als was '*' match a path
                     if ( $expr[ $y-1 ] <= self::MATCH_FILE ) {
                         $expr[ $y-1 ] = self::MATCH_PATH;
                     } else {
                         $expr[ $y++ ] = self::MATCH_FILE;
                     }
                 } else {
                     $expr[ $y++ ]=$buff[ $x ];
                 }
             }
         }
      }
      // Must match end at the end
      $expr[ $y ] = self::MATCH_THEEND;
      return $expr;
   }

   // -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* -->
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Match a pattern against a string and isolates wildcard replacement into a
    * Stack.
    *
    * @param string $inData The string to match.
    * @param string $inPattern The pattern against which to map.
    * @retval boolean true or false on the match success.
    * @exception MatcherException If data or pattern is null.
    */

   public function match( $inData, $inPattern )
   {
      $globalsModule = Environment::$modules[ 'global' ];

      $preparedPattern = $this->compilePattern( $inPattern );
      // Prime hash map
      $this->gHashMap = array(); //_pad( array(), 1, 0 );
      
      if ( strlen( $inData ) == 0 ) {
         throw new MatcherException( "No data provided" );
      }
      if ( empty( $preparedPattern ) ) {
         throw new MatcherException( "No pattern expression provided" );
      }
      $this->gLogger->debug( "Data = '$inData'" );
      // $this->gLogger->debug( "inExpr = '$preparedPattern'" );

      $buff = $this->toCharArray( $inData );
      // Allocate the result buffer
      $rslt = "";  // To stop the checker whinging about uninitialised variable
      unset( $rslt );
      
      // The previous and current position of the expression character ( MATCH_* )
      $charpos = 0;
      
      // The position in the expression, input, translation and result arrays
      $exprpos = 0;
      $buffpos = 0;
      $rsltpos = 0;
      $offset = -1;
      
      // The matching count
      $mcount = 0;
      
      // We want the complete data be in [0]
      $this->gHashMap[ $mcount ] = $inData;
      $globalsModule->add( "__0", $inData );
      
      // First check for MATCH_BEGIN
      $matchBegin = false;
      if ( $preparedPattern[ $charpos ] == self::MATCH_BEGIN ) {
         $this->gLogger->debug( "Complete data = {$this->gHashMap[ $mcount ]}" );
         $matchBegin = true;
         $exprpos = ++$charpos;
      }
   
      //Search the fist expression character ( except MATCH_BEGIN - already skipped )
      while ( $preparedPattern[ $charpos ] >= 0 ) {
         $charpos++;
      }

      //The expression character ( MATCH_* )
      $exprchr = $preparedPattern[ $charpos ];

      while ( true ) {
         /* for ( $i = 0; $i < count( $buff ); $i++ ) {
            $chr = $buff[ $i ];
            echo "buff[ $i ] : '$chr'";
            if ( $i == $buffpos ) echo "&lt== buffpos";
            echo "<br/>";
         }
         for ( $i = 0; $i < count( $preparedPattern ); $i++ ) {
            $chr = $preparedPattern[ $i ];
            echo "inExpr[ $i ] : '$chr'";
            if ( $i == $exprpos ) echo "&lt== exprpos";
            echo "<br/>";
         } */

         // Check if the data in the expression array before the current
         // expression character matches the data in the input buffer
         if ( $matchBegin ) {
            if ( !$this->matchArray( $preparedPattern, $exprpos, $charpos, $buff, $buffpos ) ) return false;
            $matchBegin = false;
         } else {
            $offset = $this->indexOfArray( $preparedPattern, $exprpos, $charpos, $buff, $buffpos );
            if ( $offset < 0 ) return false;
         }
         // Check for MATCH_BEGIN
         if ( $matchBegin ) {
            if ( $offset != 0 ) return false;
            $matchBegin = false;
         }

         // Advance buffpos
         $buffpos += ( $charpos - $exprpos );

         // Check for END's
         if ( $exprchr == self::MATCH_END ) {
            if ( $rsltpos > 0 && $this->gHashMap != NULL ) {
               // Convert the result back to a string (Java does this in a variant
               // of substr so must do it explicitly)
               $scrap = implode( "", $rslt );
               $scrap = substr( $scrap, 0, $rsltpos );
               $this->gLogger->debug( "Adding '$scrap' to map at $mcount" );
               $this->gHashMap[ ++$mcount ] = substr( $rslt, 0, $rsltpos ); // array, offset, count
               $globalsModule->add( "__$mcount", substr( $rslt, 0, $rsltpos ) );
            }
            // Don't care about rest of input buffer
            return true;
         } else if ( $exprchr == self::MATCH_THEEND ) {
            if ( $rsltpos > 0 && !empty( $this->gHashMap ) ) {
               $scrap = implode( "", $rslt );
               $scrap = substr( $scrap, 0, $rsltpos );
               $this->gLogger->debug( "Adding '$scrap' to map at $mcount" );
               $this->gHashMap[ ++$mcount ] = substr( $rslt, 0, $rsltpos );
               $globalsModule->add( "__$mcount", substr( $rslt, 0, $rsltpos ) );
            }
            // Check that we reach buffer's end
            return ( $buffpos == count( $buff ) );
         }
   
         // Search the next expression character
         $exprpos = ++$charpos;
         while ( $preparedPattern[ $charpos ] >= 0 ) $charpos++;
         $prevchr = $exprchr;
         $exprchr = $preparedPattern[ $charpos ];

         //We have here prevchr == * or **.
         $offset = ( $prevchr == self::MATCH_FILE ) ?
            $this->indexOfArray( $preparedPattern, $exprpos, $charpos, $buff, $buffpos ) :
            $this->lastIndexOfArray ( $preparedPattern, $exprpos, $charpos, $buff,$buffpos );

         if ( $offset < 0 ) return false;

         // Copy the data from the source buffer into the result buffer to substitute the expression character
        if ( $prevchr == self::MATCH_PATH ) {
             while ( $buffpos < $offset ) {
                $ch = $buff[ $buffpos++ ];
                $rslt[ $rsltpos++ ] = $ch;
               $this->gLogger->debug( "Adding '$ch' to result at {$buffpos}" );
             }
         } else {
             //Matching file, don't copy '/'
            while ( $buffpos < $offset ) {
               if ( $buff[ $buffpos ] == '/' ) return false;
               $ch = $buff[ $buffpos++ ];
               $rslt[ $rsltpos++ ] = $ch;
               $this->gLogger->debug( "Adding '$ch' to result at {$buffpos}" );
            }
         }

         if ( !isset( $rslt ) ) {
            throw new MatcherException( "No pattern expression provided in sitemap expression" );
         }

         if ( $this->gHashMap != NULL ) {
            $scrap = implode( "", $rslt );
            $scrap = substr( $scrap, 0, $rsltpos );
            $this->gLogger->debug( "Adding  $scrap to map at $mcount" );
            $this->gHashMap[ ++$mcount ] = $scrap;
            $globalsModule->add( "__$mcount", $scrap );
        }
         $rsltpos = 0;
      }
   }
   
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Get the hash map of this matcher.
    *
    * @retval string the hash map.
    */
    
   public function getHashMap()
   {
      return $this->gHashMap;
   }

   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Get the offset of a part of an int array within a char array.
    * <br>
    * This method return the index in d of the first occurrence after dpos of
    * that part of array specified by r, starting at rpos and terminating at
    * rend.
    *
    * @param array $r The array containing the data that need to be matched in d.
    * @param int $rpos The index of the first character in r to look for.
    * @param int $rend The index of the last character in r to look for plus 1.
    * @param array $d The array of char that should contain a part of r.
    * @param int $dpos The starting offset in d for the matching.
    * @retval int The offset in d of the part of r matched in d or -1 if that was
    *         not found.
    */

   private function indexOfArray( $r, $rpos, $rend, $d, $dpos )
   {
      //Check if pos and len are legal
      if ( $rend < $rpos ) throw new MatcherException( "rend < rpos, must be greater" );
      // If we need to match a zero length string return current dpos
      if ( $rend == $rpos ) return ( count( $d ) ); //?? dpos?
      // If we need to match a 1 char length string do it simply
      if ( ( $rend - $rpos) == 1 ) {
         // Search for the specified character
         for ( $x = $dpos; $x < count( $d ); $x++ ) {
            if ( $r[ $rpos ] == $d[ $x ] ) return $x;
         }
      }
      // Main string matching loop. It gets executed if the characters to
      // match are less then the characters left in the d buffer
      while ( ( $dpos + $rend - $rpos ) <= count( $d ) ) {
         // Set current startpoint in d
         $y = $dpos;
         // Check every character in d for equity. If the string is matched return dpos
         for ( $x = $rpos; $x <= $rend; $x++ ) {
            if ( $x == $rend ) return ( $dpos );
            if ( $r[ $x ] != $d[ $y++ ] ) break;
         }
         // Increase dpos to search for the same string at next offset
         $dpos++;
      }
      // The remaining chars in d buffer were not enough or the string
      // wasn't matched
      return -1;
   }

   // -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* -->
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Get the offset of a last occurance of an int array within a char array.
    * <br>
    * This method return the index in d of the last occurrence after dpos of
    * that part of array specified by r, starting at rpos and terminating at
    * rend.
    *
    * @param array $r The array containing the data that need to be matched in d.
    * @param int $rpos The index of the first character in r to look for.
    * @param int $rend The index of the last character in r to look for plus 1.
    * @param array $d The array of char that should contain a part of r.
    * @param int $dpos The starting offset in d for the matching.
    * @retval int The offset in d of the part of r matched in d or -1 if that was
    *         not found.
    */
    
    private function lastIndexOfArray( $r, $rpos, $rend, $d, $dpos )
    {
      //Check if pos and len are legal
      if ( $rend < $rpos ) throw new MatcherException( "rend < rpos, must be greater" );
        // If we need to match a zero length string return current dpos
       // If we need to match a zero length string return current dpos
      if ( $rend == $rpos ) return ( count( $d ) ); //?? dpos?

        // If we need to match a 1 char length string do it simply
        if ( ( $rend - $rpos ) == 1 ) {
            // Search for the specified character
            for ( $x = count( $d ) - 1; $x > $dpos; $x-- ) {
                if ( $r[ $rpos ] == $d[ $x ] ) return ( $x );
            }
        }

        //Main string matching loop. It gets executed if the characters to
        //match are less then the characters left in the d buffer
        $l = count( $d ) - ( $rend - $rpos );
        while ( $l >= $dpos ) {
            //Set current startpoint in d
            $y = $l;
            //Check every character in d for equity. If the string is matched
            //return dpos
            for ( $x = $rpos; $x <= $rend; $x++ ) {
                if ( $x == $rend ) return $l;
                if ( $r[ $x ] != $d[ $y++ ] ) break;
            }
            // Decrease l to search for the same string at next offset
            $l--;
        }
        // The remaining chars in d buffer were not enough or the string
        // wasn't matched
        return -1;
    }
    
   // -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* -->
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Matches elements of array (inSubArray) from rpos to rend with array d, starting from dpos.
    * <br>
    * This method return true if elements of array r from rpos to rend
    * equals elements of array d starting from dpos to dpos+(rend-rpos).
    *
    * @param array $r The array containing the data that need to be matched in d.
    * @param int $rpos The index of the first character in r to look for.
    * @param int $rend The index of the last character in r to look for plus 1.
    * @param array $d The array of char that should contain a part of r.
    * @param int $dpos The starting offset in d for the matching.
    * @retval int The offset in d of the part of r matched in d or -1 if that was
    *         not found.
    */
    
   private function matchArray( $r, $rpos, $rend, $d, $dpos )
   {
      /* for ( $i = 0; $i < count( $r ); $i++ ) {
         $chr = $r[ $i ];
         echo "r[ $i ] : '$chr'";
         if ( $i == $rpos ) echo " &lt== rpos";
         if ( $i == $rend ) echo " &lt== rend";
         echo "<br/>";
      }
      for ( $i = 0; $i < count( $d ); $i++ ) {
         $chr = $d[ $i ];
         echo "d[ $i ] : '$chr'";
         if ( $i == $dpos ) echo " &lt== dpos";
         echo "<br/>";
      } */
      //If not enough characters to match then no match
      if ( count( $d ) - $dpos < $rend - $rpos ) return false;
      for ( $i = $rpos; $i < $rend; $i++ ) {
         // echo "[1:{$r[ $i ]}][2:{$d[ $dpos++ ]}]";
         if ( $r[ $i ] != $d[ $dpos++ ] ) return false;
      }
      return true;
   }

   // -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* -->
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Converts a string to an array of characters 
    * 
    * @param string $inString string to convert
    * @retval array character array of string
    */

   function  toCharArray( $inString )
   { 
      $len = strlen( $inString ); 
      for ( $j = 0; $j < strlen ( $inString ); $j++ ) { 
         $char[ $j ]  =  substr( $inString,  $j,  1 );         
      } 
      return $char; 
   } 
   
  //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
   /**
    * Return a string representation of this class (used for debugging).
    *
    * @retval string the string representation.
    */
    
   public function toString( )
   {
      $mess = "<matcher name='{$this->gType}' src='{$this->gSrc}' package='{$this->classname( )}'/>\n";
      return $mess;
   }

}
?>
